/*	reduce.c

	AFAPI associative reduction function definitions for
	TTL_PAPERS.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"

/*	p_reduceAnd() functions...
*/

p_1u
p_reduceAnd1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAnd##suffix(register p_##suffix datum) \
{ \
	register int bits = ((8 * sizeof(datum)) - 4); \
 \
	/* Four bits at a time, and = not nand */ \
	do { \
		datum &= ~(p__nand(datum >> bits) << bits); \
	} while ((bits -= 4) >= 0); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)

#undef	P_DECLARE


p_bar
p_reduceAndbar(register p_bar datum)
{
	register int bits = ((NPROC - 1) & ~0x3);

	/* Four bits at a time, and = not nand */
	do {
		datum &= ~(p__nand(datum >> bits) << bits);
	} while ((bits -= 4) >= 0);

	return(datum);
}


p_ptr
p_reduceAndptr(register p_ptr datum)
{
	return((p_ptr) p_reduceAnd32u((p_32u) datum));
}

/*	p_reduceOr() functions...
*/

inline p_1u
p_reduceOr1u(register p_1u datum)
{
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceOr##suffix(register p_##suffix datum) \
{ \
	register int bits = ((8 * sizeof(datum)) - 4); \
 \
	/* Four bits at a time, or = nand not */ \
	do { \
		datum |= (p__nand(~(datum >> bits)) << bits); \
	} while ((bits -= 4) >= 0); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)

#undef	P_DECLARE


inline p_bar
p_reduceOrbar(register p_bar datum)
{
	register int bits = ((NPROC - 1) & ~0x3);

	/* Four bits at a time */
	do {
		datum |= (p__nand(~(datum >> bits)) << bits);
	} while ((bits -= 4) >= 0);

	return(datum);
}


p_ptr
p_reduceOrptr(register p_ptr datum)
{
	return((p_ptr) p_reduceOr32u((p_32u) datum));
}

/*	p_reduceXor() functions...

	Xor is kinda interesting because it is bounded by Or, which is
	to say that only 1 bits in the Or could be 1 bits in the Xor.
	Thus, we can compute the Or and then vote on potential 1
	bits...  this does not always result in a faster operation,
	but it will if as few as just two bits in the Or are 0s....
	Even if not, the additional cost is minimal.
*/

p_1u
p_reduceXor1u(register p_1u datum)
{
	return(p_count(datum) & 1);
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceXor##suffix(register p_##suffix datum) \
{ \
	register p_##suffix result = p_reduceOr##suffix(datum); \
	register p_##suffix bit = (((p_##suffix) 1) << \
				   ((8 * sizeof(datum)) - 1)); \
 \
	do { \
		if (bit & result) { \
			/* A potential 1 bit... 0 if even parity */ \
			if (!(p_count((datum & bit) != 0) & 1)) { \
				result ^= bit; \
			} \
		} \
 \
		bit <<= 1; \
	} while (bit); \
 \
	return(result); \
}

P_DECLARE(8u)
P_DECLARE(16u)
P_DECLARE(32u)
P_DECLARE(64u)

#undef	P_DECLARE


inline p_bar
p_reduceXorbar(register p_bar datum)
{
	register p_bar result = p_reduceOrbar(datum);
	register p_bar bit = 1;
	register int i = NPROC;

	do {
		if (bit & result) {
			/* A potential 1 bit... 0 if even parity */
			if (!(p_count((datum & bit) != 0) & 1)) {
				result ^= bit;
			}
		}

		bit <<= 1;
	} while (--i);

	return(result);
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceXor##suffix(register p_##suffix datum) \
{ \
	return((p_##suffix) p_reduceXor##suffix##u((p_##suffix##u) datum)); \
}

P_DECLARE(8)
P_DECLARE(16)
P_DECLARE(32)
P_DECLARE(64)

#undef	P_DECLARE


p_ptr
p_reduceXorptr(register p_ptr datum)
{
	return((p_ptr) p_reduceXor32u((p_32u) datum));
}

/*	p_reduceMin() functions...

	Most of these functions use an interesting voting technique in
	which each 4-bit NAND yields two bits of the minimum value, no
	matter how many processors are participating.  Signed integers
	are converted to unsigned by adding a bias factor, then the
	minimum is converted back.  We also take advantage of the fact
	that using integer compares to compare floating point values
	works for many floating point formats, thus the integer voting
	algorithm still works.  This is really a systolic algorithm
	adapted and extended for TTL_PAPERS-style NANDing.
*/

p_1u
p_reduceMin1u(register p_1u datum)
{
	return(p__all(datum));
}


static int p__xlateout[4] = {
	(0xf ^ 0x1), (0xf ^ 0x2), (0xf ^ 0x4), (0xf ^ 0x8)
};

static int p__lowbitpos[16] = {
/*	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	 0, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
};

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	register int bitpos = (8 * sizeof(datum)); \
	register p_##suffix result = 0; \
 \
	do { \
		register char mybits; \
		register char orbits; \
 \
		bitpos -= 2; \
		mybits = ((datum >> bitpos) & 3); \
		orbits = p__lowbitpos[ p__nand(p__xlateout[mybits]) ]; \
		result = ((result << 2) | orbits); \
 \
		if (orbits != mybits) { \
			/* It wasn't me, \
			   make me lose all future votes... \
			*/ \
			datum = -1; \
		} \
	} while (bitpos); \
	return(result); \
}

P_DECLARE(8u)
P_DECLARE(16u)
P_DECLARE(32u)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE
#define	P_DECLARE(suffix, bias) \
p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	return(((p_##suffix) p_reduceMin##suffix##u(datum + bias)) - bias); \
}

P_DECLARE(8, (((p_32u) 1) << 7))
P_DECLARE(16, (((p_32u) 1) << 15))
P_DECLARE(32, (((p_64u) 1) << 31))
P_DECLARE(64, (((p_64u) 1) << 63))

#undef	P_DECLARE

#ifndef	P_USE_FLOAT_COMPARE

p_32f
p_reduceMin32f(register p_32f datum)
{
	p_32f result = datum;
	*((p_32 *) &result) = p_reduceMin32(*((p_32 *) &result));
	return(result);
}

p_64f
p_reduceMin64f(register p_64f datum)
{
	p_64f result = datum;
	*((p_64 *) &result) = p_reduceMin64(*((p_64 *) &result));
	return(result);
}

#else	/* if P_USE_FLOAT_COMPARE */

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix min = datum; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
		} else { \
			register p_##suffix t = p_bcastGet##suffix(); \
			if (t < min) min = t; \
		} \
	} while (who > 0); \
 \
	return(min); \
}

P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

#endif


p_ptr
p_reduceMinptr(register p_ptr datum)
{
	return((p_ptr) p_reduceMin32u((p_32u) datum));
}

/*	p_reduceMax() functions...

	This is essentially the same as the p_reduceMin() stuff.
*/

p_1u
p_reduceMax1u(register p_1u datum)
{
	return(p__any(datum));
}


static int p__highbitpos[16] = {
/*	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
};

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	register int bitpos = (8 * sizeof(datum)); \
	register p_##suffix result = 0; \
 \
	do { \
		register char mybits; \
		register char orbits; \
 \
		bitpos -= 2; \
		mybits = ((datum >> bitpos) & 3); \
		orbits = p__highbitpos[ p__nand(p__xlateout[mybits]) ]; \
		result = ((result << 2) | orbits); \
 \
		if (orbits != mybits) { \
			/* It wasn't me, \
			   make me lose all future votes... \
			*/ \
			datum = 0; \
		} \
	} while (bitpos); \
	return(result); \
}

P_DECLARE(8u)
P_DECLARE(16u)
P_DECLARE(32u)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE
#define	P_DECLARE(suffix, bias) \
p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	return(((p_##suffix) p_reduceMax##suffix##u(datum + bias)) - bias); \
}

P_DECLARE(8, (((p_32u) 1) << 7))
P_DECLARE(16, (((p_32u) 1) << 15))
P_DECLARE(32, (((p_64u) 1) << 31))
P_DECLARE(64, (((p_64u) 1) << 63))

#undef	P_DECLARE

#ifndef	P_USE_FLOAT_COMPARE

p_32f
p_reduceMax32f(register p_32f datum)
{
	p_32f result = datum;
	*((p_32 *) &result) = p_reduceMax32(*((p_32 *) &result));
	return(result);
}

p_64f
p_reduceMax64f(register p_64f datum)
{
	p_64f result = datum;
	*((p_64 *) &result) = p_reduceMax64(*((p_64 *) &result));
	return(result);
}

#else	/* if P_USE_FLOAT_COMPARE */

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix max = datum; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
		} else { \
			register p_##suffix t = p_bcastGet##suffix(); \
			if (t > max) max = t; \
		} \
	} while (who > 0); \
 \
	return(max); \
}

P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

#endif


p_ptr
p_reduceMaxptr(register p_ptr datum)
{
	return((p_ptr) p_reduceMax32u((p_32u) datum));
}

/*	p_reduceAdd() functions...

	Slow and dumb...  TTL_PAPERS just isn't good at addition.
*/

p_1u
p_reduceAdd1u(register p_1u datum)
{
	/* Result is just 0 or 1 */
	p__any(datum);
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAdd##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix sum = 0; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
			sum += datum; \
		} else { \
			sum += p_bcastGet##suffix(); \
		} \
	} while (who > 0); \
 \
	return(sum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMul() functions...

	Slow and dumb...  TTL_PAPERS can't multiply.
*/

p_1u
p_reduceMul1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMul##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix product = 1; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
			product *= datum; \
		} else { \
			product *= p_bcastGet##suffix(); \
		} \
	} while (who > 0); \
 \
	return(product); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE
